//
// proxy
//
// proxy/forward arbitrary tcp ports to other arbitrary tcp ports on other
// hosts, checking that the client is allowed to connect using an
// authorisation file listing either hosts or regular expressions matching
// host names or equivalent dotted internet addresses
//
// copyright (c) 1998 Chris Lightfoot <chris@ex-parrot.com>
//
// version 0.1
//
// this code is covered by the GNU public license of any version you choose
//
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <regex.h>

#define PROGNAME	"proxy"
#define VERSION		"0.1"

#define COPYRIGHT	"copyright (c) 1998 Chris Lightfoot <chris@ex-parrot.com>"

// strategy is to listen on one socket, then connect on another when that
// socket is connected
int s_listen;
int s_target;
int s_client;

struct in_addr client_host;
struct sockaddr_in local_address, target_address, client_address;

char *target_name;

#define MAXPEERNAME	256
char client_name[256];

// the numeric name of the client
char client_numname[20];

struct in_addr target_host;
int local_port, target_port;

int atoport(char *service, char *proto)
{
	int port;
	long int lport;
	struct servent *serv;
	char *errpos;

	// first try to read it from /etc/services
	serv = getservbyname(service, proto);

	if (serv != NULL)
		port = serv->s_port;

	// try it as a number
	else
	{
		lport = strtol(service,&errpos,0);

		if ( (errpos[0] != 0) || (lport < 1) || (lport > 65535) )
			return -1;

		port = htons(lport);
	}

	return port;
}

struct in_addr *atoaddr(char *address)
{
	struct hostent *host;
	static struct in_addr saddr;

	// first try it as aaa.bbb.ccc.ddd
	saddr.s_addr = inet_addr(address);
	if (saddr.s_addr != -1)
		return &saddr;

	// now as a symbolic name
	host = gethostbyname(address);

	if (host != NULL)
		return (struct in_addr *) *host->h_addr_list;

	return NULL;
}

// read in a file of host specs (either dotted internet addresses, host names,
// or regexps to match one of the above)
struct hostmatch
{
#define HOST	1
#define REGEXP	2
	int type;

	struct in_addr address;
	regex_t regexp;

	struct hostmatch *next;
};

struct hostmatch *hm;

#define FILEBUFFERLEN	1024

int read_authfile(char *filename)
{
	char *buffer, *p, *q;
	struct hostmatch *h;
	struct in_addr *addr;
	regex_t rx;
	FILE *fp;
	int i, j;

	buffer = (char*)malloc(FILEBUFFERLEN);

	if (!buffer)
	{
		perror(PROGNAME": malloc");
		return -1;
	}

	fp = fopen(filename, "rt");

	if (!fp)
	{
		perror(PROGNAME": fopen");
		return -1;
	}

	i = 0;

	while (!feof(fp))
	{
		i++;
		fgets(buffer, FILEBUFFERLEN, fp);

		for (p = buffer; *p && isspace(*p); p++);

		if (*p && *p != '#')
		{
			for (q = p; *q && !isspace(*q) && *q != '#'; q++);

			*q = 0;

			if (*p == '/')
			{
				if (*(q - 1) == '/')
				{
					p++; q--; *q = 0;

					j = regcomp(&rx, p, REG_EXTENDED | REG_NOSUB | REG_ICASE);

					if (j) fprintf(stderr, PROGNAME": %s: badly formed regexp near line %d\n", filename, i);

					else
					{
						h = (struct hostmatch*)malloc(sizeof(struct hostmatch));

						if (!h)
						{
							perror(PROGNAME": malloc");
							return -1;
						}

						h->type = REGEXP;

						memcpy(&(h->regexp), &rx, sizeof(rx));

						h->next = hm;
						hm = h;
					}

				}
				else fprintf(stderr, PROGNAME": %s: badly formed regexp near line %d\n", filename, i);
			}
			else
			{
				addr = atoaddr(p);

				if (!addr)
					fprintf(stderr, PROGNAME": %s: unknown host near line %d\n", filename, i);
				else
				{
					h = (struct hostmatch*)malloc(sizeof(struct hostmatch));

					if (!h)
					{
						perror(PROGNAME": malloc");
						return -1;
					}

					h->type = HOST;

					memcpy(&(h->address), atoaddr(p), sizeof(h->address));

					h->next = hm;
					hm = h;
				}
			}
		}
	}

	fclose(fp);

	return 0;
}

// check the client host against the auth entries
int check_auth()
{
	struct hostmatch *h;

	for (h = hm; h; h = h->next)
	{
		if (h->type == HOST)
		{
			if (!memcmp(&client_host, &(h->address), sizeof(client_host)))
				return 1;
		}
		else if (h->type == REGEXP)
		{
			if (!regexec(&(h->regexp), client_name, 0, NULL, 0) ||
				!regexec(&(h->regexp), client_numname, 0, NULL, 0))
				return 1;
		}
		// can't happen
		else ;
	}

	return 0;
}

void printhelp()
{
	fprintf(stderr,
PROGNAME": proxy a tcp service from one machine to another\n"
"\n"
"usage: "PROGNAME"\n"
"       -h, --help          display this message\n"
"       -V, --version       diplay the version number ("VERSION")\n"
"       -v, --verbose       be verbose\n"
"       -l, --local-port n  listen on local port n\n"
"       -t, --target host:port\n"
"                           target to connect to\n"
"       -a, --allow file    file listing hosts allowed to use proxy\n"
"       -d, --deny file     file listing hosts not allowed to use proxy\n"
"                            (only one of -a, -d may be given)\n"
"\n"
COPYRIGHT"\n");
}

#define ARG(a, b, c) (!strcmp(a, b) || !strcmp(a, c))

#define BUFFERLEN	8192

#define AUTHNULL	0
#define AUTHALLOW	1
#define AUTHDENY	2

int main(int argc, char **argv)
{
	char **s, *t;

	int verbosity = 0;
	int authmode = AUTHNULL;

	struct in_addr *ia;

	struct hostent *he;

	int reuse_addr = 1;
	int l, m;

	fd_set rd_set;//, wr_set;
	struct timeval timeout; 

	char *x_buffer;

	pid_t child;

	struct sigaction sig;

	// ignore SIGPIPE in case of error on sockets
	sig.sa_handler = SIG_IGN;
	sig.sa_flags = 0;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGPIPE, &sig, NULL);

	// ignore SIGCHLD to avoid zombie processes
	sig.sa_handler = SIG_IGN;
#ifdef SA_NOCLDWAIT
	sig.sa_flags = SA_NOCLDWAIT;
#else
	sig.sa_flags = 0;
#endif // SA_NOCLDWAIT
	sigemptyset(&sig.sa_mask);
	sigaction(SIGCHLD, &sig, NULL);

	if (argc == 1)
	{
		fprintf(stderr, PROGNAME": you must supply arguments\n");
		printhelp();
		return 1;
	}

	// process arguments
	for (s = (argv + 1); s < (argv + argc); s++)
	{
		// help etc.
		if (ARG(*s, "-h", "--help"))
		{
			printhelp();
			return 0;
		}

		// requests for program version
		if (ARG(*s, "-V", "--version"))
		{
			printf(PROGNAME": version "VERSION"\n\n"COPYRIGHT"\n");
			return 0;
		}

		// verbosity
		if (ARG(*s, "-v", "--verbose"))
			verbosity = 1;

		// request for the local port to use
		if (ARG(*s, "-l", "--local-port"))
		{
			if (local_port)
			{
				fprintf(stderr, PROGNAME": you may only specify one local port\n");
				return 1;
			}

			else
			{
				local_port = atoport(*(++s), "tcp");

				if (local_port == -1)
				{
					fprintf(stderr, PROGNAME": the local port was not resolved\n");
					return 1;
				}
				if (local_port < 1024)
				{
					fprintf(stderr, PROGNAME": the local port must be unprivileged\n");
					return 1;
				}
			}
		}

		// the target host/port to use
		if (ARG(*s, "-t", "--target"))
		{
			if (target_port || target_name)
			{
				fprintf(stderr, PROGNAME": you may only specify one target\n");
				return 1;
			}

			else
			{
				t = strchr(*(++s), ':');

				if (!t)
				{
					fprintf(stderr, PROGNAME": target should be specified as host:port\n");
					return 1;
				}

				*t++ = 0;

				target_name = (char*)malloc(strlen(*s) + 1);
				if (!target_name)
				{
					perror(PROGNAME": malloc");
					return 1;
				}

				strcpy(target_name, *s);

				// try to resolve the target port
				target_port = atoport(t, "tcp");
				if (target_port == -1)
				{
					fprintf(stderr, PROGNAME": the target port was not resolved\n");
					return 1;
				}

				// now try to resolve the target name
				ia = atoaddr(*s);

				if (!ia)
				{
					fprintf(stderr, PROGNAME": the target host was not resolved\n");
					return 1;
				}

				memcpy(&target_host, ia, sizeof(struct in_addr));
			}
		}

		if (ARG(*s, "-a", "--allow"))
		{
			if (authmode != AUTHNULL)
			{
				fprintf(stderr, PROGNAME": you may only specify a single -a or -d option\n");
				return 1;
			}

			if (read_authfile(*(++s)) < 0)
				return 1;

			authmode = AUTHALLOW;
		}

		if (ARG(*s, "-d", "--deny"))
		{
			if (authmode != AUTHNULL)
			{
				fprintf(stderr, PROGNAME": you may only specify a single -a or -d option\n");
				return 1;
			}

			if (read_authfile(*(++s)) < 0)
				return 1;

			authmode = AUTHDENY;
		}
	}

	if (!local_port || !target_port || !target_name)
	{
		fprintf(stderr, PROGNAME": you must specify a local port and a target host and port\n");
		return 1;
	}

	// create a socket to listen on
	s_listen = socket(AF_INET, SOCK_STREAM, 0);

	if (s_listen < 0)
	{
		perror(PROGNAME": socket(s_listen)");
		return 1;
	}

	// set up the listening address
	memset(&local_address, 0, sizeof(local_address));
	local_address.sin_family = AF_INET;
	local_address.sin_port = local_port;
	local_address.sin_addr.s_addr = htonl(INADDR_ANY);

	// set up the target address
	memset(&target_address, 0, sizeof(target_address));
	target_address.sin_family = AF_INET;
	target_address.sin_port = target_port;
	memcpy(&(target_address.sin_addr), &target_host, sizeof(target_host));

	// reuse the address
	if (setsockopt(s_listen, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof(reuse_addr)) < 0)
	{
		perror(PROGNAME": setsockopt");
		return 1;
	}

	// bind to address
	if (bind(s_listen, (struct sockaddr*)&local_address, sizeof(local_address)) < 0)
	{
		perror(PROGNAME": bind");
		return 1;
	}

	// start listening
	if (listen(s_listen, SOMAXCONN) < 0)
	{
		perror(PROGNAME": listen");
		return 1;
	}

	// be garrulous
	if (verbosity)
	{
		fprintf(stderr, PROGNAME": being garrulous\n");
		fprintf(stderr, PROGNAME": listening on local port %d\n", ntohs(local_port));
		fprintf(stderr, PROGNAME": forwarding connections to %s:%d\n", target_name, ntohs(target_port));

		if (authmode == AUTHNULL)
			fprintf(stderr, PROGNAME": not using an auth file\n");
		else if (authmode == AUTHALLOW)
			fprintf(stderr, PROGNAME": allowing access to hosts in auth file\n");
		else if (authmode == AUTHDENY)
			fprintf(stderr, PROGNAME": denying access to hosts in auth file\n");
	}

	// now we get to the meat of the problem
	//
	// we listen for incoming connections, and fork off sub-processes to
	// deal with them
	for (;;)
	{
		l = sizeof(client_address);
		s_client = accept(s_listen, &client_address, &l);

		if (s_client < 0)
		{
			perror(PROGNAME": accept");
//			return 1; //???
		}

		// try to find out who is connecting
		he = gethostbyaddr((char*)&client_address.sin_addr, sizeof(client_address.sin_addr), AF_INET);

		strcpy(client_numname, inet_ntoa(client_address.sin_addr));
		memcpy(&client_host, &(client_address.sin_addr), sizeof(struct in_addr));

		if (!he)
		{
			if (verbosity)
				fprintf(stderr, PROGNAME": failed to resolve symbolic name for peer\n");
		}

		else
			strcpy(client_name, he->h_name);

		if (verbosity)
			fprintf(stderr, PROGNAME": got connection from %s\n", *client_name ? client_name : client_numname);

		child = fork();

		if (child < 0) // oops
		{
			perror(PROGNAME": fork");
//			return 1; //???
		}

		else if (child > 0) // this is still the parent
			// close our copy of connected socket
			close(s_client);

		else if (!child) // this is the child
		{
			// close our copy of listening socket
			close(s_listen);

			if (verbosity)
				fprintf(stderr, PROGNAME": forked child %d to handle connection\n", getpid());

			if (authmode != AUTHNULL)
			{
				l = check_auth();

				if (verbosity)
					fprintf(stderr, PROGNAME": %s client in authorisation list\n", l ? "found" : "did not find");

				if ((l && authmode == AUTHDENY) || (!l && authmode == AUTHALLOW))
				{
					if (verbosity)
						fprintf(stderr, PROGNAME": rejected connection from %s\n", *client_name ? client_name : client_numname);

					shutdown(s_client, 2);

					return 0;
				}
			}

			x_buffer = (char*)malloc(BUFFERLEN);

			if (!x_buffer)
			{
				perror(PROGNAME": malloc");
				return 1;
			}

	
			// create also a socket to connect on
			s_target = socket(AF_INET, SOCK_STREAM, 0);

			if (s_target < 0)
			{
				perror(PROGNAME": socket(s_target)");
				return 1;
			}

			// now we connect to the remote host
			if (connect(s_target, &target_address, sizeof(target_address)) < 0)
			{
				perror(PROGNAME": connect");
				return 1;
			}

			for (;;)
			{
				// we go into a loop select'ing on the two sockets
				FD_SET(s_client, &rd_set);
				FD_SET(s_target, &rd_set);

				timeout.tv_sec = 0;
				timeout.tv_usec = 5000000; // 5s

				if (select((s_client > s_target ?
						s_client : s_target) + 1,
						&rd_set, NULL,
						NULL, &timeout))
				{
					// can we read from client and write
					// to target?
					if (FD_ISSET(s_client, &rd_set))
					{
						l = read(s_client, x_buffer, BUFFERLEN);

						if (l < 0 && errno != EINTR)
						{
							perror(PROGNAME": read(client)");
							return 1;
						}

						else if (!l)
						{
							if (verbosity) fprintf(stderr, PROGNAME": closed connection from %s\n", *client_name ? client_name : client_numname);

							if (shutdown(s_client, 2) < 0)
								perror(PROGNAME": shutdown(client)");
							if (shutdown(s_target, 2) < 0)
								perror(PROGNAME": shutdown(target)");

							if (verbosity) fprintf(stderr, PROGNAME": child %d exiting\n", getpid());

							return 0;
						}

						for (t = x_buffer; l > 0;)
						{
							m = write(s_target, t, l);

							if (m > 0)
							{
								t += m;
								l -= m;
							}
							else if (errno != EINTR)
							{
								perror(PROGNAME": write(target)");
								return 1; //???
							}
						}
					}

					// can we read from target and write
					// to client?
					if (FD_ISSET(s_target, &rd_set))
					{
						l = read(s_target, x_buffer, BUFFERLEN);

						if (l < 0 && errno != EINTR)
						{
							perror(PROGNAME": read(target)");
							return 1;
						}
						else if (!l)
						{
							if (verbosity) fprintf(stderr, PROGNAME": closed connection from %s\n", *client_name ? client_name : client_numname);

							if (shutdown(s_client, 2) < 0)
								perror(PROGNAME": shutdown(client)");
							if (shutdown(s_target, 2) < 0)
								perror(PROGNAME": shutdown(target)");


							if (verbosity) fprintf(stderr, PROGNAME": child %d exiting\n", getpid());
							return 0;
						}

						for (t = x_buffer; l > 0;)
						{
							m = write(s_client, t, l);

							if (m > 0)
							{
								t += m;
								l -= m;
							}
							else if (errno != EINTR)
							{
								perror(PROGNAME": write(client)");
								return 1; //???
							}
						}
					}
				}
			}
		}
	}
}
